/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.se.jsr181;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.axis2.AxisFault;
import org.apache.axis2.deployment.util.Utils;
import org.apache.axis2.description.AxisOperation;
import org.apache.axis2.description.WSDL2Constants;
import org.apache.axis2.description.java2wsdl.AnnotationConstants;
import org.apache.axis2.engine.MessageReceiver;
import org.apache.axis2.util.Loader;
import org.codehaus.jam.JAnnotation;
import org.codehaus.jam.JClass;
import org.codehaus.jam.JMethod;
import org.codehaus.jam.JamClassIterator;
import org.codehaus.jam.JamService;
import org.codehaus.jam.JamServiceFactory;
import org.codehaus.jam.JamServiceParams;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;

/**
 * Deployer used to create Axis Services
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 1.0
 * 
 */
public class Axis2Deployer {

    /**
     * The Axis2 configuration context
     */
    private Component component;

    private Logger logger;

    /**
     * Creates a new instance of {@link Axis2Deployer}
     * 
     * @param context
     */
    public Axis2Deployer(final Component component) {
        this.component = component;
        this.logger = component.getLogger();
    }

    /**
     * Deploy a JAXWS service i.e. create a new instance of the service from the
     * annotated class
     * 
     * @param className
     */
    public AxisService deployJAXWS(final String className, final ClassLoader classloader,
            final Provides provides, final URL serviceLocation) {
        AxisService axisService = null;
        try {
            JamServiceFactory factory = JamServiceFactory.getInstance();
            JamServiceParams jamServiceParams = factory.createServiceParams();
            jamServiceParams.addClassLoader(Thread.currentThread().getContextClassLoader());
            jamServiceParams.includeClass(className);
            JamService service = factory.createService(jamServiceParams);
            JamClassIterator jClassIter = service.getClasses();
            while (jClassIter.hasNext()) {
                JClass jclass = (JClass) jClassIter.next();
                if (jclass.getQualifiedName().equals(className)) {
                    /**
                     * Schema genertaion done in two stage 1. Load all the
                     * methods and create type for methods parameters (if the
                     * parameters are Bean then it will create Complex types for
                     * those , and if the parameters are simple type which
                     * decribe in SimpleTypeTable nothing will happen) 2. In the
                     * next stage for all the methods messages and port types
                     * will be creteated
                     */
                    JAnnotation annotation = jclass.getAnnotation(AnnotationConstants.WEB_SERVICE);
                    if (annotation != null) {
                        // try to see whether JAX-WS jars in the class path , if
                        // so
                        // use them
                        // to process annotated pojo else use annogen to process
                        // the
                        // pojo class
                        axisService = createAxisService(className, serviceLocation, classloader,
                                provides);

                        fixMEP(axisService, jclass);

                        this.component.getAxisContext().getAxisConfiguration().addService(
                                axisService);
                    } else {
                        axisService = createAxisServiceUsingAnnogen(className, classloader,
                                serviceLocation, provides);
                        this.component.getAxisContext().getAxisConfiguration().addService(
                                axisService);
                    }
                }
            }
            // FIXME : Handle exceptions
        } catch (AxisFault e) {
            this.logger.warning(e.getMessage());
        } catch (MalformedURLException e) {
            this.logger.warning(e.getMessage());
        } catch (IOException e) {
            this.logger.warning(e.getMessage());
        } catch (ClassNotFoundException e) {
            this.logger.warning(e.getMessage());
        } catch (InstantiationException e) {
            this.logger.warning(e.getMessage());
        } catch (IllegalAccessException e) {
            this.logger.warning(e.getMessage());
        }

        return axisService;
    }

    /**
     * Deploy a service from an archive
     * 
     * @param archive
     */
    public void deployArchive(File archive) {
        // TODO
    }

    /**
     * Create an Axis Service
     * 
     * @param className
     * @param serviceLocation
     * @param classLoader
     * @return
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     * @throws AxisFault
     */
    private AxisService createAxisService(String className, URL serviceLocation,
            ClassLoader classLoader, Provides provides) throws ClassNotFoundException,
            InstantiationException, IllegalAccessException, AxisFault {
        org.apache.axis2.description.AxisService axisService;
        AxisService petalsService = null;
        try {
            Class<?> claxx = Class.forName("org.apache.axis2.jaxws.description.DescriptionFactory");
            Method mthod = claxx.getMethod("createAxisService", new Class[] { Class.class });
            Class<?> pojoClass = Loader.loadClass(classLoader, className);

            // FIXME :
            //            initServiceObject(pojoClass);

            axisService = (org.apache.axis2.description.AxisService) mthod.invoke(claxx,
                    new Object[] { pojoClass });

            // delegate
            petalsService = new AxisService(axisService, provides);

            if (axisService != null) {
                Iterator operations = petalsService.getOperations();
                while (operations.hasNext()) {
                    AxisOperation axisOperation = (AxisOperation) operations.next();
                    if (axisOperation.getMessageReceiver() == null) {
                        try {
                            Class jaxwsMR = Loader
                                    .loadClass("org.apache.axis2.jaxws.server.JAXWSMessageReceiver");
                            MessageReceiver jaxwsMRInstance = (MessageReceiver) jaxwsMR
                                    .newInstance();
                            axisOperation.setMessageReceiver(jaxwsMRInstance);
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                    }
                }
            }
            petalsService.setElementFormDefault(false);
            petalsService.setFileName(serviceLocation);
            Utils.fillAxisService(petalsService, this.component.getAxisContext()
                    .getAxisConfiguration(), new ArrayList(), new ArrayList());
            // Not needed at this case, the message receivers always set to RPC
            // if this
            // executes
            // setMessageReceivers(axisService);

        } catch (Exception e) {
            this.logger.warning(e.getMessage());
            // Seems like the jax-ws jars missing in the class path .
            // lets try with annogen
            petalsService = createAxisServiceUsingAnnogen(className, classLoader, serviceLocation,
                    provides);
        }
        return petalsService;
    }

    /**
     * Create an Axis Service
     * 
     * @param className
     * @param classLoader
     * @param serviceLocation
     * @return
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     * @throws AxisFault
     */
    private AxisService createAxisServiceUsingAnnogen(String className, ClassLoader classLoader,
            URL serviceLocation, Provides provides) throws ClassNotFoundException,
            InstantiationException, IllegalAccessException, AxisFault {
        Map<String, MessageReceiver> messageReceiverMap = new HashMap<String, MessageReceiver>();

        // The INONLY message receiver
        Class<?> inOnlyMessageReceiver = Loader
                .loadClass("org.apache.axis2.rpc.receivers.RPCInOnlyMessageReceiver");
        MessageReceiver messageReceiver = (MessageReceiver) inOnlyMessageReceiver.newInstance();
        messageReceiverMap.put(WSDL2Constants.MEP_URI_IN_ONLY, messageReceiver);

        // The INOUT message receiver
        Class<?> inoutMessageReceiver = Loader
                .loadClass("org.apache.axis2.rpc.receivers.RPCMessageReceiver");
        MessageReceiver inOutmessageReceiver = (MessageReceiver) inoutMessageReceiver.newInstance();
        messageReceiverMap.put(WSDL2Constants.MEP_URI_IN_OUT, inOutmessageReceiver);
        messageReceiverMap.put(WSDL2Constants.MEP_URI_ROBUST_IN_ONLY, inOutmessageReceiver);

        // create the Axis2 service
        org.apache.axis2.description.AxisService axisService = org.apache.axis2.description.AxisService
                .createService(className, component.getAxisContext().getAxisConfiguration(),
                        messageReceiverMap, null, null, classLoader);

        // TODO : get the POJO and call the init(Component)

        // create the PEtALS Axis Service
        AxisService petalsService = new AxisService(axisService, provides);
        petalsService.setFileName(serviceLocation);
        return petalsService;
    }

    /**
     * Fix MEP in Axis Operations from java method declarations. The Axis2
     * builder does not set the good MEPs...
     */
    private void fixMEP(AxisService axisService, JClass jclass) {
        JMethod[] methods = jclass.getMethods();
        for (int i = 0; i < methods.length; i++) {
            JMethod jMethod = methods[i];
            try {
                AxisOperation tmpOperation = Utils.getAxisOperationForJmethod(jMethod);
                AxisOperation operation = axisService.getOperation(tmpOperation.getName());
                if (operation != null) {
                    if (this.logger.isLoggable(Level.FINE)) {
                        this.logger.log(Level.FINE, "Setting MEP to '"
                                + tmpOperation.getMessageExchangePattern() + "' on operation : "
                                + operation.getName() + "(Original MEP was '"
                                + operation.getMessageExchangePattern() + "')");
                    }
                    operation.setMessageExchangePattern(tmpOperation.getMessageExchangePattern());
                } else {
                    // no such operation in service...
                    this.logger.log(Level.FINE, "Operation not found for method : "
                            + jMethod.getSimpleName());
                }
            } catch (AxisFault e) {
                this.logger.log(Level.WARNING, "Can not get the Axis Operation for method "
                        + jMethod.getSimpleName());
            }
        }
    }

    /**
     * Add the component in the POJO service FIXME : Do some injection in the
     * POJO
     * 
     * @param clazz
     */
    private void initServiceObject(Class<?> clazz) {
        Method method = null;
        try {
            method = clazz.getMethod("init",
                    new Class[] { org.ow2.petals.component.framework.api.Component.class });
        } catch (Exception e) {
            logger.log(Level.FINE, "Can not find the good init method");
        }

        if (method != null) {
            try {
                method.invoke(clazz, new Object[] { component });
            } catch (IllegalAccessException e) {
                logger.log(Level.WARNING, "Exception trying to call init method", e);
            } catch (IllegalArgumentException e) {
                logger.log(Level.WARNING, "Exception trying to call init method", e);
            } catch (InvocationTargetException e) {
                logger.log(Level.WARNING, "Exception trying to call init method", e);
            }
        }
    }
}
